/*  Screem:  screem-tagtree.c
 *
 *  the tag tree object
 *
 *  Copyright (C) 2001 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */
#include <config.h>

#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <gtk/gtktreeview.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkhbox.h>

#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>

#include "xml.h"

#include "screem-tagtree.h"

#include "screem-window.h"
#include "screem-window-private.h"
#include "screem-editor.h"

/* FIXME: make properties of the ScreemTagTree */
static GList *g_name = NULL;  /* group name */
static GList *g_node = NULL;  /* group node */

typedef struct NodeInfo {
	gchar *name;
	gchar *open;
	gchar *close;
} NodeInfo;

enum {
	ARG_0,
	ARG_WINDOW
};

static void 
screem_tag_tree_add_node( ScreemTagTree *tree,
			  xmlNodePtr node, GtkTreeIter *parent );

static void add_bluefish_fn( ScreemTagTree *tree );


static void screem_tag_tree_clear_branch( ScreemTagTree *tree,
					  GtkTreePath *path,
					  gboolean siblings );

static void screem_tag_tree_class_init( ScreemTagTreeClass *klass );
static void screem_tag_tree_init( ScreemTagTree *tag_tree );
static void screem_tag_tree_finalize( GObject *object );
static void screem_tag_tree_set_prop( GObject *object, guint property_id,
				    const GValue *value, GParamSpec *pspec );
static void screem_tag_tree_get_prop( GObject *object, guint property_id,
				      GValue *value, GParamSpec *pspec);
static void screem_tag_tree_size_request( GtkWidget *widget, GtkRequisition *req );
static void screem_tag_tree_realize( GtkWidget *widget );
static void screem_tag_tree_show( GtkWidget *widget );

ScreemTagTree *screem_tag_tree_new()
{
	ScreemTagTree *tree;
	GType type;

	type = screem_tag_tree_get_type();
	
	tree = SCREEM_TAG_TREE( g_object_new( type, NULL ) );

	return tree;
}

void screem_tag_tree_add_dtd( ScreemTagTree *tag_tree, ScreemDTD *dtd )
{
	xmlDocPtr tree;
	xmlNodePtr root;
	xmlNodePtr newnode;
	gchar *opentag;
	gchar *closetag;

	const GSList *elements;

	if( ! dtd ) {
		return;
	}
		
	elements = screem_dtd_get_elements( dtd );
	
	tree = xmlNewDoc( XML_DEFAULT_VERSION );
	xmlDocSetRootElement( tree, xmlNewDocNode( tree, NULL, "TagTree", NULL ) );
	root = xmlNewChild( xmlDocGetRootElement( tree ), NULL, "node", NULL );

	xmlSetProp( root, "name", "Doctype" );

	while( elements ) {
		const gchar *name;
		ScreemDTDClose close;

		name = screem_dtd_element_get_name( (ScreemDTDElement*)elements->data );
		close = screem_dtd_element_close_state( dtd, name );
		
		newnode = xmlNewChild( root, NULL, "node", NULL );
		opentag = g_strconcat( "<", name, ">", NULL );
		closetag = g_strconcat( "</", name, ">", NULL );

		xmlSetProp( newnode, "name", name );
		xmlSetProp( newnode, "open", opentag );
		if( close != MUSTNT ) {
			xmlSetProp( newnode, "close", closetag );
		}
		g_free( opentag );
		g_free( closetag );

		elements = elements->next;
	}

	screem_tag_tree_add_node( tag_tree, root, NULL );
}


/* static stuff */
static void 
screem_tag_tree_add_node( ScreemTagTree *tree, xmlNodePtr node, GtkTreeIter *parent )
{
	gchar *name;

	GtkTreeIter iter;

	g_assert( node );

	name = xmlGetProp( node, "name" );
	if( name ) {
		/* does name already match the name of a sibling node of 
		   parent ? */
		GtkTreePath *path;

		path = g_hash_table_lookup( tree->paths, name );

		/* only deal with toplevel */
		if( path && ! parent ) {
			/* erase it + remove from hash table */
			g_hash_table_remove( tree->paths, name );
		
			screem_tag_tree_clear_branch( tree, path, FALSE );

			gtk_tree_path_free( path );
		}

				
		/* add the node */
		{
			NodeInfo *info;

			gtk_tree_store_append( tree->store, &iter, parent );
			
			info = g_new0( NodeInfo, 1 );
			info->name = name;
			info->open = xmlGetProp( node, "open" );
			info->close = xmlGetProp( node, "close" );

			gtk_tree_store_set( tree->store, &iter, 0, name, -1 );
			gtk_tree_store_set( tree->store, &iter, 1, info, -1 );

			if( ! parent ) {
				path = gtk_tree_model_get_path( GTK_TREE_MODEL( tree->store )
								, &iter );
				
				g_hash_table_insert( tree->paths, name, path );
			}
		}
	}
	
	/* has children? */
	if( node->xmlChildrenNode )
		screem_tag_tree_add_node( tree,
					  node->xmlChildrenNode, &iter );
	
	/* move on to the next sibling */
	if( node->next ) {
		screem_tag_tree_add_node( tree, node->next, parent );
	}
}


static void add_bluefish_fn( ScreemTagTree *tree )
{
 const gchar *home;
        
        /* 
         * the first element in a pair is the file name, the second is
         * the group name for the tag tree
         */
        gchar *fn_files[] = {
                "php3_functions", "PHP3",
                "php4_functions", "PHP4",
                "rxml_functions", "RXML",
                "ssi_functions", "SSI",
                NULL
        };

        gchar *fn_path;
        gchar *fn_file;
        gint i;

        FILE *file;
        struct stat s;
        gchar *data;
        gint size;

        gchar *search;	
        gchar *line;

	GtkTreeIter iter;

        GList *list;

        home = g_get_home_dir();

	fn_path = g_strjoin( G_DIR_SEPARATOR_S, home, ".bluefish", NULL );

        for( i = 0; fn_files[ i ]; i += 2 ) {
                fn_file = g_strjoin( G_DIR_SEPARATOR_S, fn_path, fn_files[ i ],
                                     NULL );
                if( stat( fn_file, &s ) < 0 ) {
                        g_free( fn_file );
                        continue;
                }
                if( ! ( file = fopen( fn_file, "r" ) ) ) {
                        g_free( fn_file );
                        continue;
                }
                g_free( fn_file );
                data = g_new( gchar, s.st_size + 1 );
                size = fread( data, 1, s.st_size, file );
                fclose( file );
                data[ size ] = '\0';

                /* file now read, start processing */
	

                g_free( data );
        }

	g_list_foreach( g_name, (GFunc)g_free, NULL );
        g_list_free( g_name );
        g_list_free( g_node );
       
        g_free( fn_path );
}


static void screem_tag_tree_clear_branch( ScreemTagTree *tree,
					  GtkTreePath *path,
					  gboolean siblings )
{
	GtkTreeIter iter;
	GValue value = {0,};
	NodeInfo *info;

	if( ! path )
		return;

	if( ! gtk_tree_model_get_iter( GTK_TREE_MODEL( tree->store ), 
				       &iter, path ) )
		return;
	
	gtk_tree_model_get_value( GTK_TREE_MODEL( tree->store ),
				  &iter, 1, &value );
	info = g_value_get_pointer( &value );
	if( info ) {
		if( info->name )
			g_free( info->name );
		if( info->close )
			g_free( info->close );
		if( info->open )
			g_free( info->open );
		g_free( info );
	}
	g_value_unset( &value );

	/* descend child branchs */
	{
		GtkTreePath *child;
		gchar *str;

		child = gtk_tree_path_copy( path );
		gtk_tree_path_down( child );
		str = gtk_tree_path_to_string( child );
		g_free( str );
		screem_tag_tree_clear_branch( tree, child, TRUE );
		gtk_tree_path_free( child );
	}

	/* follow siblings? */
	if( siblings ) {
		GtkTreePath *sibling;
	
		sibling = gtk_tree_path_copy( path );
		gtk_tree_path_next( sibling );
		screem_tag_tree_clear_branch( tree, sibling, TRUE );
		gtk_tree_path_free( sibling );
	}
	
	/* remove this node */
	gtk_tree_store_remove( tree->store, &iter );
}


static gboolean screem_tag_tree_click( GtkTreeView *tree_view,
				       GtkTreePath *path,
				       GtkTreeViewColumn *column,
				       gpointer data )
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *sel;

	ScreemTagTree *tag_tree;

	tag_tree = SCREEM_TAG_TREE( data );


	model = GTK_TREE_MODEL( tag_tree->store );
	sel = gtk_tree_view_get_selection( GTK_TREE_VIEW( tree_view ) );

	if( sel ) {
		GValue value = {0,};
		NodeInfo *info;
		ScreemWindow *window;
		ScreemEditor *editor;

		window = SCREEM_WINDOW( tag_tree->window );

		editor = SCREEM_EDITOR( window->details->editor );

		gtk_tree_selection_get_selected( sel, &model, &iter );
		gtk_tree_model_get_value( model, &iter, 1, &value );

		info = g_value_get_pointer( &value );

		if( info ) {
			screem_editor_insert_markup( editor, info->open,
						     info->close );
		}
		g_value_unset( &value );
		
	} else {
		/* hmm, no iterator */
		
	}

	return (sel != NULL);
}


static gboolean screem_tag_tree_press( GtkWidget *widget, 
				       GdkEventButton *event,
				       ScreemTagTree *tag_tree )
{
	return FALSE;
}


/* G Object stuff */
#define PARENT_TYPE GTK_TYPE_HBOX

static gpointer parent_class;

static void screem_tag_tree_class_init( ScreemTagTreeClass *klass )
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	GParamSpec *pspec;

	object_class = G_OBJECT_CLASS( klass );
	widget_class = (GtkWidgetClass *)klass;

	object_class->finalize = screem_tag_tree_finalize;
	parent_class = g_type_class_peek_parent( klass );

	object_class->get_property = screem_tag_tree_get_prop;
	object_class->set_property = screem_tag_tree_set_prop;

	pspec = g_param_spec_pointer( "window", "window",
				      "window object",
				      G_PARAM_READABLE |
				      G_PARAM_WRITABLE );
	g_object_class_install_property( object_class,
					 ARG_WINDOW, pspec );

	widget_class->show = screem_tag_tree_show;
	widget_class->realize = screem_tag_tree_realize;
	widget_class->size_request = screem_tag_tree_size_request;
}

static void screem_tag_tree_init( ScreemTagTree *tag_tree )
{
	xmlNodePtr node;
	xmlDocPtr tree;
	xmlDocPtr user_doc;
	const gchar *home;
	const gchar *mime_type;
	gchar *path;
	gchar *file;
	gint num;
	gint pos;
	struct dirent **namelist;

	GtkWidget *sw;
	GtkCellRenderer *rend;
	GtkTreeViewColumn *col;

	sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
					GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
                                               GTK_SHADOW_IN );
	tag_tree->store = gtk_tree_store_new( 2, G_TYPE_STRING,
					      G_TYPE_POINTER );
	tag_tree->view = 
		gtk_tree_view_new_with_model(GTK_TREE_MODEL(tag_tree->store));
        gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( tag_tree->view ), TRUE );

	tag_tree->paths = g_hash_table_new( g_str_hash, g_str_equal );

	rend = gtk_cell_renderer_text_new ();
	col = gtk_tree_view_column_new_with_attributes ("Tag Tree",
							rend,
							"text", 0,
							NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tag_tree->view), col);

	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tag_tree->view), FALSE );

	gtk_widget_show( tag_tree->view );
	tree = tag_tree->tree = xmlParseFile( TAG_TREE_FILE );

	g_assert( GTK_IS_BOX( tag_tree ) );

	/* load user tag trees */
	home = g_get_home_dir();
	path = g_strconcat( home, G_DIR_SEPARATOR_S, ".screem", NULL );
	num = scandir( path, &namelist, 0, alphasort );
	g_free( path );

	for( pos = 0; pos < num; pos ++ ) {
		file = namelist[ pos ]->d_name;
		mime_type = gnome_vfs_get_mime_type( file );
		if( mime_type &&
		    ! strcmp( "application/x-screem-tag-tree", mime_type ) ) {
			path = g_strconcat( home, G_DIR_SEPARATOR_S, 
					    ".screem", G_DIR_SEPARATOR_S,
					    file, NULL );
			user_doc = xmlParseFile( path );
			g_free( path );
			/* copy the root node */
			node = xmlDocGetRootElement( user_doc );
			node = xmlCopyNode( node->xmlChildrenNode, TRUE );
			/* add the node to the main tree */
			xmlAddChild( xmlDocGetRootElement( tree ), node );
			/* delete user_doc */
			xmlFreeDoc( user_doc );
		}
		free( namelist[ pos ] );
	}
	if( num > 0 ) {
		free( namelist );
	}

	g_assert( GTK_IS_BOX( tag_tree ) );

	node = xmlDocGetRootElement( tree );

	/* not interested in the first element */
	node = node->xmlChildrenNode;

	g_assert( GTK_IS_BOX( tag_tree ) );

	screem_tag_tree_add_node( tag_tree, node, NULL );

	add_bluefish_fn( tag_tree );

	g_assert( GTK_IS_BOX( tag_tree ) );

	g_signal_connect( G_OBJECT( tag_tree->view ), "row_activated",
			  G_CALLBACK( screem_tag_tree_click ),
			  tag_tree );

	g_signal_connect( G_OBJECT( tag_tree->view ), "button_press_event",
			  G_CALLBACK( screem_tag_tree_press ),
			  tag_tree );

	gtk_container_add( GTK_CONTAINER( sw ), tag_tree->view );

	g_assert( GTK_IS_BOX( tag_tree ) );
	gtk_box_pack_start( GTK_BOX( tag_tree ), sw, TRUE, TRUE, 0 );

	gtk_widget_show( sw );
}


static void screem_tag_tree_finalize( GObject *object )
{
	G_OBJECT_CLASS( parent_class )->finalize( object );
}

static void screem_tag_tree_set_prop( GObject *object, guint property_id,
				    const GValue *value, GParamSpec *pspec )
{
	ScreemTagTree *tree;

	tree = SCREEM_TAG_TREE( object );

	switch( property_id ) {
	case ARG_WINDOW:
		tree->window = g_value_get_pointer( value );

		break;
	}
}

static void screem_tag_tree_get_prop( GObject *object, guint property_id,
				    GValue *value, GParamSpec *pspec)
{
	switch( property_id ) {
	case ARG_WINDOW:
		g_value_set_pointer( value,
				     (gpointer)
				     SCREEM_TAG_TREE( object )->window );
		break;
	}
}


static void screem_tag_tree_size_request( GtkWidget *widget, GtkRequisition *req )
{
	GTK_WIDGET_CLASS( parent_class )->size_request( widget, req );
}

static void screem_tag_tree_realize( GtkWidget *widget )
{
	GTK_WIDGET_CLASS( parent_class )->realize( widget );
}

static void screem_tag_tree_show( GtkWidget *widget )
{
	ScreemTagTree *tag_tree;

	tag_tree = SCREEM_TAG_TREE( widget );

	GTK_WIDGET_CLASS( parent_class )->show( widget );

	gtk_widget_show( tag_tree->view );
}

GType screem_tag_tree_get_type()
{
	static GType type = 0;
	
	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemTagTreeClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_tag_tree_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemTagTree ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_tag_tree_init
		};

		type = g_type_register_static( PARENT_TYPE,
					       "ScreemTagTree",
					       &info, 0 );
	}

	return type;
}
