/*  Screem:  screem-window-toolbars.c
 *
 *  Copyright (C) 2002 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <bonobo/bonobo-ui-util.h>
#include <bonobo/bonobo-object.h>
#include <bonobo/bonobo-ui-component.h>
#include <bonobo/bonobo-control.h>

#include <gtk/gtkcombo.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkiconfactory.h>

#include <glib/glist.h>

#include <gdk-pixbuf/gdk-pixbuf.h>

#include <libgnome/gnome-exec.h>

#include <string.h>

#include "screem-application.h"
#include "screem-window.h"
#include "screem-window-private.h"
#include "screem-editor.h"
#include "screem-todo.h"

#include "pageUI.h"

typedef enum {
	LEFT, CENTRE, JUSTIFY, RIGHT
} Align;

static void
toolbar_external_browse_callback( BonoboUIComponent *component,
				   gpointer user_data,
				   const char *verb )
{
	ScreemWindow *window;
	ScreemPage *page;

	window = SCREEM_WINDOW( user_data );
	page =  screem_window_get_document( window );

	external_view( page );
}

static void
toolbar_task_list_callback( BonoboUIComponent *component,
			    gpointer user_data,
			    const char *verb )
{
	ScreemWindow *window;
	ScreemSite *site;
	ScreemTodo *todo;

	window = SCREEM_WINDOW( user_data );
	site = screem_window_get_current( window );
	todo = screem_site_get_todo( site );

	screem_todo_view( todo );
}

static void screem_toolbar_spell_check( BonoboUIComponent *component,
					gpointer user_data,
					const char *verb )
{
	ScreemWindow *window;
	ScreemPage *page;

	window = SCREEM_WINDOW( user_data );
	page = screem_window_get_document( window );

	g_object_set_data( G_OBJECT( window->details->spell ), "page", page );
	screem_spell_check_interactive( window->details->spell );
}

static void screem_toolbar_external_browser_notify( GConfClient *client,
						    guint cnxn_id,
						    GConfEntry *entry,
						    gpointer data )
{
	ScreemApplication *application;
	ScreemWindow *window;
	BonoboUIComponent *ui;
	CORBA_Environment ev;
	GSList *list;

	gchar *browser;
	gchar *conf;
	gchar *tip;
	gint i;
	gboolean set;
	gchar *path;
	
	window = SCREEM_WINDOW( data );
	application = SCREEM_APPLICATION( window->application );
	ui = screem_window_get_shell_ui( window );

	list = window->details->browser_notify_handles;

	CORBA_exception_init( &ev );
	bonobo_ui_component_freeze( ui, &ev );

	i = g_slist_index( list, GINT_TO_POINTER( cnxn_id ) );
	i /= 2;
	i ++;

	browser = g_strdup_printf( "Browser%i", i );

	conf = g_strdup_printf( "/apps/screem/Browser%i_path", i );
	tip = gconf_client_get_string( application->client, conf,
				       NULL );
	g_free( conf );

	set = TRUE;
	if( ! tip || *tip == '\0' ) {
		set = FALSE;
		tip = g_strdup_printf( "External Browser %i Not Set",
				       i );
	}
	
	path = g_strconcat( "/commands/", browser, NULL );

	bonobo_ui_component_set_prop( ui, path, "tip", tip, &ev );
	bonobo_ui_component_set_prop( ui, path, "hidden",
					set ? "0" : "1", NULL );
					
	g_free( path );
	
	g_free( tip );
	g_free( browser );

	bonobo_ui_component_thaw( ui, &ev );
	CORBA_exception_free( &ev );
}

void screem_toolbar_external_browser( BonoboUIComponent *component,
				      gpointer user_data,
				      const char *verb )
{
	ScreemWindow *window;
	ScreemApplication *application;
	ScreemPage *page;
	const gchar *pagepath;

	gchar *conf;
	gchar *browserpath;
	gboolean use_url;
	gboolean dynamic;

	window = SCREEM_WINDOW( user_data );

	/* the verb holds the browser we want, eg Browser1, Browser2 etc */

	application = SCREEM_APPLICATION( window->application );

	/* get browser settings */
	conf = g_strconcat( "/apps/screem/", verb, "_path", NULL );
	browserpath = gconf_client_get_string( application->client,
					       conf, NULL );
	g_free( conf );

	if( ! browserpath && ! strcmp( "Browser0", verb ) ) {
		/* default browser, hasn't been overridden for
		   SCREEM, so use gnome default,
		   /desktop/gnome/applications/browser/exec */
		browserpath = gconf_client_get_string( application->client,
						       "/desktop/gnome/applications/browser/exec", NULL );
	}

	page = screem_window_get_document( window );
	dynamic = screem_page_is_dynamic( page );

	if( ! dynamic ) {
		conf = g_strconcat( "/apps/screem/", verb, "_static", NULL );
	} else {
		conf = g_strconcat( "/apps/screem/", verb, "_dynamic", NULL );
	}
	use_url = gconf_client_get_bool( application->client, conf, NULL );
	g_free( conf );

	if( ! browserpath ) {
		/* FIXME: hmm, no browserpath available, alert user */
	} else if( (screem_page_save_confirmation( page ) != 
		    GTK_RESPONSE_CANCEL) ) {
		pagepath = screem_page_get_pathname( page );

		if( use_url ) {
			ScreemSite *site;
			const gchar *sitepath;
			gchar *uri;
			
			site = screem_window_get_current( window );
			
			sitepath = screem_site_get_pathname( site );
			
			uri = g_strconcat( browserpath, " http://",
					   screem_site_get_http_url( site ),
					   pagepath + strlen( sitepath ), 
					   NULL );
			
			gnome_execute_shell( NULL, uri );
			g_free( uri );
		} else {
			gchar *file;
			gchar *rest;
			gchar *uri;
			gint pos;

			file = strstr( browserpath, "%f" );
			if( file ) {
				pos = file - browserpath;
				rest = file + 2;
				if( rest > ( browserpath + strlen( browserpath ) ) )
					rest = "";
				file = g_strndup( browserpath, pos );
			} else {
				/* no %f, assume path just goes on the end */
				file = g_strdup( browserpath );
				rest = "";
			}
			uri = g_strconcat( file, " ", pagepath, " ", rest,
					   NULL );
			g_free( file );

			gnome_execute_shell( NULL, uri );

			g_free( uri );
		}
	}
	if( browserpath )
		g_free( browserpath );
}

static void screem_toolbar_create_browsers( ScreemWindow *window )
{
	BonoboUIComponent *ui;
	CORBA_Environment ev;
	gint i;

	ScreemApplication *application;
	gchar *conf;
	const int NO_OF_CUSTOM_BROWSERS = 3;

	application = SCREEM_APPLICATION( window->application );

	CORBA_exception_init( &ev );

	ui = screem_window_get_shell_ui( window );

	bonobo_ui_component_freeze( ui, &ev );


	for( i = 1; i <= NO_OF_CUSTOM_BROWSERS; ++ i ) {
		gchar *data;
		gchar *label;
		gchar *tip;
		gchar *command;
		gchar *browser = g_strdup_printf( "Browser%i", i );
		gchar *menu = 
			g_strdup_printf( "<menuitem name=\"%s\" verb=\"\"/>",
					 browser );
		gchar *toolbar = 
			g_strdup_printf( "<toolitem name=\"%s\" verb=\"\"/>",
					 browser );
		const gchar *pixtype;
		gint handle;
		gboolean set;
		
		conf = g_strdup_printf( "/apps/screem/Browser%i_icon", i );
		data = gconf_client_get_string( application->client, conf,
						NULL );
		g_free( conf );
		if( data ) {
			/* change default stock icon */
			GdkPixbuf *pixbuf;
			gchar *id = g_strconcat( "Screem_", browser, NULL );
			GtkIconSet *set = gtk_icon_factory_lookup_default( id );
			GtkIconSource *src = gtk_icon_source_new();
				
			pixbuf = gdk_pixbuf_new_from_file( data, NULL );
			gtk_icon_source_set_pixbuf( src, pixbuf );
			
			gtk_icon_source_set_direction_wildcarded( src, TRUE );
			gtk_icon_source_set_state_wildcarded( src, TRUE );
			gtk_icon_source_set_size_wildcarded( src, TRUE );
			gtk_icon_source_set_size( src, GTK_ICON_SIZE_BUTTON );
			gtk_icon_set_add_source( set, src );
			
			g_object_unref( G_OBJECT( pixbuf ) );
			gtk_icon_source_free( src );

			g_free( data );
			data = id;
		} else {
			data = g_strconcat( "Screem_", browser, NULL );
		}

		conf = g_strdup_printf( "/apps/screem/Browser%i_path", i );
		tip = gconf_client_get_string( application->client, conf,
					       NULL );
		g_free( conf );
		set = TRUE;
		if( ! tip || *tip == '\0' ) {
			set = FALSE;
			tip = g_strdup_printf( "External Browser %i Not Set",
					       i );
		}
		label = g_strdup_printf( "External Browser %i", i );

		command = g_strdup_printf( "<cmd name=\"%s\" _tip=\"%s\" _label=\"%s\" pixtype=\"stock\" pixname=\"%s\" verb=\"%s\" />",  
					   browser, tip, label,
					   data, browser );
		
		g_free( label );
		g_free( tip );
		g_free( data );
		
		bonobo_ui_component_set_translate( ui, "/commands", command, &ev );

		bonobo_ui_component_set_translate( ui, "/Browser Toolbar", toolbar, &ev );

		bonobo_ui_component_set_translate( ui, "/menu/View", menu, &ev );

		if( ! set ) {
			gchar *path;
			
			path = g_strconcat( "/commands/", browser, NULL );
			bonobo_ui_component_set_prop( ui, path, "hidden",
							"1", NULL );
			g_free( path );
		}
		
		bonobo_ui_component_add_verb( ui, browser, 
					      screem_toolbar_external_browser,
					      window );
		g_free( command );
		g_free( browser );

		browser = g_strdup_printf( "/apps/screem/Browser%i_icon",
					   i );
		handle = gconf_client_notify_add( application->client, browser,
						  
						  screem_toolbar_external_browser_notify, 
						  window, NULL, NULL );
		window->details->browser_notify_handles =
			g_slist_append(window->details->browser_notify_handles,
				       GINT_TO_POINTER( handle ) );
		g_free( browser );

		browser = g_strdup_printf( "/apps/screem/Browser%i_path",
					   i );
		handle = gconf_client_notify_add( application->client, browser,
						  
						  screem_toolbar_external_browser_notify, 
						  window, NULL, NULL );
		window->details->browser_notify_handles =
			g_slist_append(window->details->browser_notify_handles,
				       GINT_TO_POINTER( handle ) );
		g_free( browser );

	}

	bonobo_ui_component_thaw( ui, &ev );
	CORBA_exception_free( &ev );
}


void screem_toolbar_markup( ScreemEditor *editor, const gchar *tag )
{
	gchar *open;
	gchar *close;

	open = g_strconcat( "<", tag, ">", NULL );
	close = g_strconcat( "</", tag, ">", NULL );

	screem_editor_insert_markup( editor, open, close );

	g_free( close );
	g_free( open );
}

void screem_toolbar_specific_markup( ScreemEditor *editor, const gchar *open, 
				     const gchar *close )
{
	screem_editor_insert_markup( editor, open, close );
}

void screem_toolbar_align( ScreemEditor *editor, Align a )
{
	const gchar *open = NULL;
	const gchar *close = NULL;

	switch( a ) {
	case CENTRE:
		open = "<center>";
		close = "</center>";
		break;
	case RIGHT:
		open = "<div align=\"right\">";
		close = "</div>";
		break;
	case JUSTIFY:
		open = "<div align=\"justify\">";
		close = "</div>";
		break;
	case LEFT:
		open = "<div align=\"left\">";
		close = "</div>";
		break;
	}
	if( open && close )
		screem_editor_insert_markup( editor, open, close );
}


/* normal tag insertion handling */
static void quickbar_bold_callback( BonoboUIComponent *component,
				    gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "b" );
}
static void quickbar_italic_callback( BonoboUIComponent *component,
				      gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "i" );
}
static void quickbar_underline_callback( BonoboUIComponent *component,
					 gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "u" );
}
static void quickbar_para_callback( BonoboUIComponent *component,
				    gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "p" );
}
static void quickbar_bullet_callback( BonoboUIComponent *component,
				      gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "ul" );
}
static void quickbar_numbered_callback( BonoboUIComponent *component,
					gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "ol" );
}

/* align handling */

static void quickbar_alignleft_callback( BonoboUIComponent *component,
					 gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_align( editor, LEFT );
}
static void quickbar_aligncentre_callback( BonoboUIComponent *component,
					   gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_align( editor, CENTRE );
}
static void quickbar_alignright_callback( BonoboUIComponent *component,
					  gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_align( editor, RIGHT );
}
static void quickbar_justify_callback( BonoboUIComponent *component,
				       gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_align( editor, JUSTIFY );
}


static void fontbar_font_callback( BonoboUIComponent *component,
				   gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "font" );
}
static void fontbar_pre_callback( BonoboUIComponent *component,
				  gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "pre" );
}
static void fontbar_subscript_callback( BonoboUIComponent *component,
					gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "sub" );
}

static void fontbar_superscript_callback( BonoboUIComponent *component,
					  gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "sup" );
}



static void tablebar_table_callback( BonoboUIComponent *component,
				     gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "table" );
}

static void tablebar_tablebody_callback( BonoboUIComponent *component,
					 gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "tbody" );
}

static void tablebar_tr_callback( BonoboUIComponent *component,
				  gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "tr" );
}

static void tablebar_th_callback( BonoboUIComponent *component,
				  gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "th" );
}

static void tablebar_td_callback( BonoboUIComponent *component,
				  gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "td" );
}

static void tablebar_caption_callback( BonoboUIComponent *component,
				       gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "caption" );
}

static void tablebar_summary_callback( BonoboUIComponent *component,
				       gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "summary" );
}


static void formbar_textarea_callback( BonoboUIComponent *component,
				       gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "textarea" );	
}

static void formbar_input_callback( BonoboUIComponent *component,
				    gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_specific_markup( editor, "<input name=\"\">", NULL );
}

static void formbar_fileinput_callback( BonoboUIComponent *component,
					gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_specific_markup( editor, "<input type=\"file\" name=\"\">",
					NULL );
}

static void formbar_button_callback( BonoboUIComponent *component,
				     gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_markup( editor, "button" );
}

static void formbar_checkbutton_callback( BonoboUIComponent *component,
					  gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_specific_markup( editor, 
					"<input type=\"checkbox\" name=\"\">",
					NULL );
}

static void formbar_radiobutton_callback( BonoboUIComponent *component,
					  gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_specific_markup( editor, 
					"<input type=\"radio\" name=\"\">", NULL );
}

static void formbar_optionmenu_callback( BonoboUIComponent *component,
					 gpointer user_data, const char *verb )
{
	ScreemEditor *editor;
	editor = SCREEM_EDITOR( SCREEM_WINDOW( user_data )->details->editor );
	screem_toolbar_specific_markup( editor, 
					"<select><option>",
					"</option></select>" );
}

void screem_window_initialise_toolbars( ScreemWindow *window )
{
	BonoboUIVerb verbs[] = {
		BONOBO_UI_VERB( "externalBrowse", 
				toolbar_external_browse_callback ),
		BONOBO_UI_VERB( "TaskList", toolbar_task_list_callback ),

		BONOBO_UI_VERB( "Browser0", screem_toolbar_external_browser ),

		BONOBO_UI_VERB( "SpellCheck", screem_toolbar_spell_check ),

		/* quickbar */
		BONOBO_UI_VERB( "Bold", quickbar_bold_callback ),
		BONOBO_UI_VERB( "Italic", quickbar_italic_callback ),
		BONOBO_UI_VERB( "Underline", quickbar_underline_callback ),
		BONOBO_UI_VERB( "Para", quickbar_para_callback ),
		BONOBO_UI_VERB( "Bullet", quickbar_bullet_callback ),
		BONOBO_UI_VERB( "Numbered", quickbar_numbered_callback ),
		BONOBO_UI_VERB( "AlignLeft", quickbar_alignleft_callback ),
		BONOBO_UI_VERB( "AlignCentre", quickbar_aligncentre_callback ),
		BONOBO_UI_VERB( "AlignRight", quickbar_alignright_callback ),
		BONOBO_UI_VERB( "Justify", quickbar_justify_callback ),

		/* fonts */
		BONOBO_UI_VERB( "Font", fontbar_font_callback ),
		BONOBO_UI_VERB( "Pre", fontbar_pre_callback ),
		BONOBO_UI_VERB( "Subscript", fontbar_subscript_callback ),
		BONOBO_UI_VERB( "Superscript", fontbar_superscript_callback ),

		/* tables */
		BONOBO_UI_VERB( "Table", tablebar_table_callback ),
		BONOBO_UI_VERB( "TBody", tablebar_tablebody_callback ),
		BONOBO_UI_VERB( "Tr", tablebar_tr_callback ),
		BONOBO_UI_VERB( "Th", tablebar_th_callback ),
		BONOBO_UI_VERB( "Td", tablebar_td_callback ),
		BONOBO_UI_VERB( "Caption", tablebar_caption_callback ),
		BONOBO_UI_VERB( "Summary", tablebar_summary_callback ),

		/* forms */
		BONOBO_UI_VERB( "Textarea", formbar_textarea_callback ),
		BONOBO_UI_VERB( "Input", formbar_input_callback ),
		BONOBO_UI_VERB( "Fileinput", formbar_fileinput_callback ),
		BONOBO_UI_VERB( "Button", formbar_button_callback ),
		BONOBO_UI_VERB( "Checkbutton", formbar_checkbutton_callback ),
		BONOBO_UI_VERB( "Radiobutton", formbar_radiobutton_callback ),
		BONOBO_UI_VERB( "Optionmenu", formbar_optionmenu_callback ),

		/* helpers */

		BONOBO_UI_VERB_END
	};

	screem_toolbar_create_browsers( window );

	bonobo_ui_component_freeze( window->details->shell_ui, NULL );

	bonobo_ui_component_add_verb_list_with_data( window->details->shell_ui,
						     verbs, window );

	bonobo_ui_component_thaw( window->details->shell_ui, NULL );
}
