/*  Screem:  screem-window.c
 *
 *  The ScreemWindow widget
 *
 *  Copyright (C) 2001 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */
#include <config.h>

#include <string.h>

#include <bonobo/bonobo-ui-util.h>
#include <bonobo/bonobo-control.h>
#include <bonobo/bonobo-ui-component.h>

#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gnome-popup-menu.h>

#include <libgnomevfs/gnome-vfs-uri.h>

#include <gtk/gtkhbox.h>
#include <gtk/gtknotebook.h>

#include <gdk/gdktypes.h>

#include <gmodule.h>

#include <glade/glade.h>

#include <gtk/gtktreeview.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkhpaned.h>
#include <gtk/gtkvbox.h>

#include <gtk/gtkeventbox.h>

#include <gconf/gconf-client.h>

#include "screem-window.h"

#include "screem-window-private.h"

#include "screem-editor.h"
#include "screem-preview.h"
#include "screem-site.h"
#include "screem-site-view.h"
#include "screem-tagtree.h"
#include "screem-linkview.h"
#include "screem-tree-view.h"

#include "support.h"

#include "pageUI.h"

#include "fileops.h"

#include "rb-sidebar.h"
#include "rb-sidebar-button.h"

enum {
	ARG_0,
	ARG_APP_ID,
	ARG_APP
};

typedef enum _Drop_types {
	TARGET_URI_LIST,
	TARGET_URL,
	TARGET_MOZ
} Drop_types;

static const GtkTargetEntry drop_types[] = {
        { "text/uri-list", 0, TARGET_URI_LIST },
	{ "text/x-moz-url", 0, TARGET_MOZ },
        { "_NETSCAPE_URL", 0, TARGET_URL },
        { "x-url/http", 0, TARGET_URL },
        { "x-url/ftp", 0, TARGET_URL },
};
static const gint num_drop_types = sizeof(drop_types) / sizeof(drop_types [0]);

static void screem_window_initialise_class( ScreemWindowClass *klass );
static void screem_window_initialise( ScreemWindow *window );
static void screem_window_finalize( GObject *object );
static void screem_window_set_prop( GObject *object, guint prop_id, 
				    const GValue *value, GParamSpec *spec );
static void screem_window_get_prop( GObject *object, guint prop_id, 
				    GValue *value, GParamSpec *spec );
static void screem_window_size_request( GtkWidget *widget,
					GtkRequisition *req );
static void screem_window_realize( GtkWidget *widget );
static void screem_window_show( GtkWidget *widget );



void screem_window_set_document( ScreemWindow *window, ScreemPage *page );


static void screem_window_constructed( ScreemWindow *window );

static gboolean screem_window_file_view_click( GtkTreeView *tree_view,
					       GtkTreePath *path,
					       GtkTreeViewColumn *column,
					       gpointer data );

static gboolean screem_window_file_view_press( GtkWidget *widget, 
					       GdkEventButton *event,
					       ScreemWindow *window );

static void screem_window_tab_clicked( GtkWidget *label, GdkEventButton *event,
                                       ScreemWindow *window );
static void screem_window_tab_popup_switch( GtkWidget *widget, gpointer data );

static void screem_window_notebook_change( GtkNotebook *notebook, 
					   GtkNotebookPage *p,
                                           guint num );

static void screem_window_real_show_message( ScreemWindow *window,
					     GtkWidget *view, 
					     const gchar *message );
static void screem_window_real_clear_messages( GtkWidget *view );
static gboolean screem_window_clear_status( ScreemWindow *window );


static void screem_window_highlight_spelling( ScreemSpell *spell,
					      gint start, gint end,
					      ScreemWindow *window );

static void screem_window_shortcut_notify( GConfClient *client,
					   guint cnxn_id,
					   GConfEntry *entry,
					   gpointer data );

static void screem_window_set_page_label( ScreemPage *page,
					  gboolean status,
					  gpointer data );

static void screem_window_set_doctype( ScreemPage *page,
				       const gchar *doctype,
				       gpointer data );

static void screem_window_select_context_callback( ScreemTreeView *view, 
						   guint start,
						   guint end, gpointer data );
static void screem_window_request_context_callback( ScreemTreeView *view, 
						    guint pos,
						    gchar **context_name,
						    guint *depth,
						    guint *start,
						    guint *end,
						    gboolean build,
						    gpointer data );

static void screem_window_view_change( RBSidebarButton *button, 
				       ScreemWindow *window );
static void
screem_window_switch_site_destroy( gpointer data, GClosure *closure );
static void
screem_window_switch_site( BonoboUIComponent *uic, 
			   gpointer data, const char *cname );

static void screem_window_site_switch( RBSidebarButton *button,
				       SwitchSiteData *sd );

static void screem_window_drop( GtkWidget *widget, GdkDragContext *context,
				gint x, gint y, 
				GtkSelectionData *selectionData,
				guint info, guint time, ScreemWindow *window );

void
screem_window_close( ScreemWindow *window )
{
	ScreemSite *site;
	ScreemWindowDetails *details;
	gint pos;
	gint w;
	gint h;

	g_return_if_fail( SCREEM_IS_WINDOW( window ) );

	details = window->details;

	site = screem_window_get_current( window );

       	details->documents = NULL;

	if( GTK_WIDGET_VISIBLE( window->details->pane ) ) {
		pos = gtk_paned_get_position( GTK_PANED( window->details->pane ) );
		gconf_client_set_int( window->application->client,
				      "/apps/screem/ui/barsize", pos, NULL );
	}

	gtk_window_get_size( GTK_WINDOW( window ), &w, &h );
	gconf_client_set_int( window->application->client,
			      "/apps/screem/ui/window_width", w, NULL );
	gconf_client_set_int( window->application->client,
			      "/apps/screem/ui/window_height", h, NULL );

	details->current = NULL;
	if( details->current_document ) {
		g_signal_handlers_disconnect_matched(G_OBJECT( details->current_document ),
						     G_SIGNAL_MATCH_FUNC |
						     G_SIGNAL_MATCH_DATA,
						     0, 0, NULL,
						     screem_window_set_doctype,
						     window );
		g_signal_handlers_disconnect_matched( G_OBJECT( details->current_document ),
						G_SIGNAL_MATCH_DATA,
						0, 0, NULL, NULL,
						details->tree_view );
		g_signal_handlers_disconnect_matched( G_OBJECT( details->current_document ),
						G_SIGNAL_MATCH_DATA,
						0, 0, NULL, NULL,
						details->editor );
		g_signal_handlers_disconnect_matched( G_OBJECT( details->current_document ),
						G_SIGNAL_MATCH_DATA,
						0, 0, NULL, NULL,
						details->preview );
		g_signal_handlers_disconnect_matched( G_OBJECT( details->current_document ),
						G_SIGNAL_MATCH_DATA,
						0, 0, NULL, NULL,
						details->link_view );
	}
	details->current_document = NULL;

	gtk_widget_destroy( GTK_WIDGET( window ) );
}

static void
screem_window_constructed( ScreemWindow *window )
{
	GtkWidget *pane;
	GtkWidget *box;
	GtkWidget *sbox;
	GtkWidget *image;
	GtkWidget *label;
	GladeXML *xml;

	gchar *glade_path;

	ScreemWindowDetails *details;

	BonoboControl *control;

	gint barsize;

	GtkWidget *button;
       	gboolean offline;
       
	details = window->details;

	gtk_window_set_wmclass( GTK_WINDOW( window ), "Screem", 
				"mainwindow"  );

	/* Bonobo setup */
	details->ui_container = bonobo_window_get_ui_container( BONOBO_WINDOW( window ) );
	

	/* load ui from xml file */
	window->details->shell_ui = bonobo_ui_component_new( "Screem Window" );
	bonobo_ui_component_set_container( details->shell_ui,
					   screem_window_get_ui_container(window),
					   NULL );
	bonobo_ui_component_freeze( details->shell_ui, NULL );
	bonobo_ui_engine_config_set_path( bonobo_window_get_ui_engine( BONOBO_WINDOW( window ) ), "/apps/screem/UIConfig/kvps" );
	bonobo_ui_util_set_ui( details->shell_ui,
			       UIDATADIR,
			       "screem-window-ui.xml",
			       "screem", NULL );

	/* content_hbox is the main display area of the window */
	window->content_hbox = gtk_hbox_new( FALSE, 0 );

	details->pane = pane = gtk_hpaned_new();
	gtk_box_pack_start( GTK_BOX( window->content_hbox ), pane,
			    TRUE, TRUE, 0 );

	details->barbox = sbox = gtk_vbox_new( FALSE, 0 );
	gtk_widget_show( sbox );
	gtk_paned_add1( GTK_PANED( pane ), sbox );

	button = gtk_button_new_with_label( _( "Views" ) );
	gtk_widget_show_all( button );
	gtk_box_pack_start( GTK_BOX( sbox ), button, FALSE, FALSE, 0 );
       
	details->bar = rb_sidebar_new();
	gtk_widget_show( details->bar );
	gtk_box_pack_start( GTK_BOX( sbox ), details->bar, TRUE, TRUE, 0 );

	details->sitebar = rb_sidebar_new();

	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( gtk_widget_show_all ),
			       details->bar, NULL, G_CONNECT_SWAPPED );
	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( gtk_widget_hide ),
			       details->sitebar, NULL, G_CONNECT_SWAPPED );

	button = gtk_button_new_with_label( _( "Sites" ) );
	gtk_widget_show_all( button );
	gtk_box_pack_start( GTK_BOX( sbox ), button, FALSE, FALSE, 0 );

	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( gtk_widget_show_all ),
			       details->sitebar, NULL, G_CONNECT_SWAPPED );
	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( gtk_widget_hide ),
			       details->bar, NULL, G_CONNECT_SWAPPED );

	gtk_box_pack_start( GTK_BOX( sbox ), details->sitebar, TRUE, TRUE, 0 );

	barsize = gconf_client_get_int( window->application->client,
					"/apps/screem/ui/barsize", NULL );
	gtk_paned_set_position( GTK_PANED( pane ), barsize );

	details->bar_notify = gconf_client_notify_add( window->application->client,
						       "/apps/screem/ui/show_shortcut",
						       screem_window_shortcut_notify,
						       window, NULL, NULL );

	/* create file tree */
	details->file_view = gtk_notebook_new();
        gtk_notebook_set_tab_pos( GTK_NOTEBOOK( details->file_view ),
                                  GTK_POS_BOTTOM );

        gtk_widget_show( details->file_view );


	/* create site tab */
	label = gtk_label_new( _( "Site" ) );
        gtk_widget_show( label );
	image = gtk_image_new_from_stock( "gtk-home",
					  GTK_ICON_SIZE_SMALL_TOOLBAR );
	{
		GtkWidget *tree;
		GtkWidget *scrolled;

		tree = gtk_tree_view_new();
		scrolled = gtk_scrolled_window_new( NULL, NULL );
		gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( scrolled ),
						GTK_POLICY_AUTOMATIC, 
						GTK_POLICY_AUTOMATIC );
		gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( scrolled ),
							GTK_SHADOW_IN );
		gtk_container_add( GTK_CONTAINER( scrolled ), tree );

		screem_site_init_tree_view( tree );
		g_object_set_data( G_OBJECT( window ), "siteview", tree );

		g_signal_connect( G_OBJECT( tree ), "row_activated",
				  G_CALLBACK( screem_window_file_view_click ),
				  window );
		g_signal_connect( G_OBJECT( tree ), "button_press_event",
				  G_CALLBACK( screem_window_file_view_press ),
				  window );
		
		gtk_widget_show_all( scrolled );
		
		screem_window_add_side_tab( window, label, image, scrolled );
	}

	/* create resources tree */
	details->tagtree = GTK_WIDGET( screem_tag_tree_new() );
	g_object_ref( G_OBJECT( details->tagtree ) );
	gtk_object_sink( GTK_OBJECT( details->tagtree ) );
	g_object_set( G_OBJECT( details->tagtree ), "window",
		      window, NULL );
	gtk_widget_show( details->tagtree );
	label = gtk_label_new( _( "Site" ) );
        gtk_widget_show( label );
	image = gtk_image_new_from_stock( "Screem_Resources",
					  GTK_ICON_SIZE_SMALL_TOOLBAR );
	screem_window_add_side_tab( window, label, image, details->tagtree );


	/* add the editor / preview / link view etc */
	details->editor = GTK_WIDGET( screem_editor_new( window ) );
	g_object_ref( G_OBJECT( details->editor ) );
	gtk_object_sink( GTK_OBJECT( details->editor ) );
	gtk_widget_show_all( details->editor );
	details->current = details->editor;
	
	g_signal_connect( G_OBJECT( details->editor ),
			  "request_context",
			  G_CALLBACK( screem_window_request_context_callback ),
			  window );
	button = GTK_WIDGET(rb_sidebar_button_new( "editor", _( "Editor" ) ) );
	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( screem_window_view_change ),
			       window, NULL, 0 );
	rb_sidebar_button_set( RB_SIDEBAR_BUTTON( button ), "Screem_Site",
			       _( "Editor" ), TRUE );
       	rb_sidebar_append( RB_SIDEBAR( details->bar ), 
			   RB_SIDEBAR_BUTTON( button ) );

	details->preview = 
		GTK_WIDGET( screem_preview_new_with_window( window ) );
	g_object_ref( G_OBJECT( details->preview ) );
	gtk_object_sink( GTK_OBJECT( details->preview ) );
	gtk_widget_show_all( details->preview );

	button = GTK_WIDGET( rb_sidebar_button_new( "preview",
						    _( "Preview" ) ) );
	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( screem_window_view_change ),
			       window, NULL, 0 );
	rb_sidebar_button_set( RB_SIDEBAR_BUTTON( button ), "Screem_Site",
			       _( "Preview" ), TRUE );
       	rb_sidebar_append( RB_SIDEBAR( details->bar ), 
			   RB_SIDEBAR_BUTTON( button ) );

	details->link_view = GTK_WIDGET( screem_link_view_new( window ) );
	g_object_ref( G_OBJECT( details->link_view ) );
	gtk_object_sink( GTK_OBJECT( details->link_view ) );
	gtk_widget_show_all( details->link_view );
	button = GTK_WIDGET( rb_sidebar_button_new( "linkview", 
						    _( "Link View" ) ) );
	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( screem_window_view_change ),
			       window, NULL, 0 );
	rb_sidebar_button_set( RB_SIDEBAR_BUTTON( button ), "Screem_LinkView",
			       _( "LinkView" ), TRUE );
       	rb_sidebar_append( RB_SIDEBAR( details->bar ), 
			   RB_SIDEBAR_BUTTON( button ) );	
	
	details->tree_view = GTK_WIDGET( screem_tree_view_new( window ) );
	g_object_ref( G_OBJECT( details->tree_view ) );
	gtk_object_sink( GTK_OBJECT( details->tree_view ) );
	gtk_widget_show_all( details->tree_view );
	
	g_signal_connect( G_OBJECT( details->tree_view ),
			  "select_context",
			  G_CALLBACK( screem_window_select_context_callback ),
			  window );
	button = GTK_WIDGET( rb_sidebar_button_new( "treeview", 
						    _( "Tree View" ) ) );
	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( screem_window_view_change ),
			       window, NULL, 0 );
	rb_sidebar_button_set( RB_SIDEBAR_BUTTON( button ), "Screem_Site",
			       _( "Tree View" ), TRUE );
       	rb_sidebar_append( RB_SIDEBAR( details->bar ), 
			   RB_SIDEBAR_BUTTON( button ) );	
	
	gtk_widget_show_all( details->bar );

	box = gtk_vbox_new( FALSE, 0 );

	/* add all views into a box */
	details->box = gtk_hbox_new( FALSE, 0 );
	gtk_box_pack_start( GTK_BOX( details->box ),
			    details->editor, TRUE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX( details->box ),
			    details->preview, TRUE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX( details->box ),
			    details->link_view, TRUE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX( details->box ),
			    details->tree_view, TRUE, TRUE, 0 );

	gtk_widget_hide( details->editor );
	gtk_widget_hide( details->preview );
	gtk_widget_hide( details->link_view );
	gtk_widget_hide( details->tree_view );
	
	gtk_box_pack_start( GTK_BOX( box ), details->box, TRUE, TRUE, 0 );

	gtk_widget_show( box );
	gtk_paned_add2( GTK_PANED( pane ), box );

	control = bonobo_control_new( details->file_view );

	bonobo_ui_component_object_set( details->shell_ui, 
					"/File List/files",
					BONOBO_OBJREF( control ), NULL );
	bonobo_object_unref( BONOBO_OBJECT( control ) );

	/* create messages control */
	glade_path = gconf_client_get_string( window->application->client,
					      "/apps/screem/general/glade_path",
					      NULL );
	xml = glade_xml_new( glade_path, "messages_book", NULL );
	g_free( glade_path );
	details->messages = glade_xml_get_widget( xml, "messages_book" );

	control = bonobo_control_new( details->messages );
	bonobo_ui_component_object_set( details->shell_ui, 
					"/Messages/messages",
					BONOBO_OBJREF( control ), NULL );
	bonobo_object_unref( BONOBO_OBJECT( control ) );


	/* now create the notebook for the page tabs */
	details->notebook = gtk_notebook_new();
	gtk_notebook_set_scrollable( GTK_NOTEBOOK( details->notebook ), TRUE );
	gtk_notebook_set_tab_pos( GTK_NOTEBOOK( details->notebook ),
				  GTK_POS_BOTTOM );
	gtk_notebook_set_show_border( GTK_NOTEBOOK( details->notebook ), FALSE );
	gtk_notebook_set_tab_border( GTK_NOTEBOOK( details->notebook ), 0 );
	gtk_box_pack_start( GTK_BOX( box ), details->notebook, FALSE, FALSE, 0 );
	g_signal_connect( G_OBJECT( details->notebook ), "switch_page",
			  G_CALLBACK( screem_window_notebook_change ),
			  NULL );
	g_object_set_data( G_OBJECT( details->notebook ), "window", window );
	gtk_widget_show( details->notebook );

	gtk_widget_show( pane );

	gtk_widget_show( details->box );
	gtk_widget_show( details->current );
	gtk_widget_show( window->content_hbox );

	bonobo_window_set_contents( BONOBO_WINDOW( window ),
				    window->content_hbox );

	screem_window_initialise_menus( window );
	screem_window_initialise_toolbars( window );

	/* add recent pages and sites menu */
	details->recent_pages = 
		gedit_recent_new_with_ui_component( "screem-pages", 4,
						    details->shell_ui,
						    "/menu/File/RecentPages" );
	details->recent_sites =
		gedit_recent_new_with_ui_component( "screem-sites", 4,
						    details->shell_ui,
						    "/menu/Site/RecentSites" );
	g_signal_connect( G_OBJECT( details->recent_pages ),
			  "activate",
			  G_CALLBACK( screem_window_recent_page_cb ),
			  (gpointer)window );
	g_signal_connect( G_OBJECT( details->recent_sites ),
			  "activate",
			  G_CALLBACK( screem_window_recent_site_cb ),
			  (gpointer)window );

	/* add online status control to status bar */
	offline = window->details->offline = window->application->offline;

	button = gtk_button_new();
	gtk_button_set_relief( GTK_BUTTON( button ),
			       GTK_RELIEF_NONE );
	details->online_status = gtk_image_new();
	gtk_container_add( GTK_CONTAINER( button ), details->online_status );
	gtk_widget_show_all( button );
	g_signal_connect_data( G_OBJECT( button ), "clicked",
			       G_CALLBACK( screem_window_toggle_online_status),
			       window, NULL, G_CONNECT_SWAPPED );

	control = bonobo_control_new( button );
	bonobo_ui_component_object_set( details->shell_ui, 
					"/status/onlineStatus",
					BONOBO_OBJREF( control ), NULL );
	bonobo_object_unref( BONOBO_OBJECT( control ) );

	gtk_widget_set_size_request( button, -1, 16 );
	
	screem_window_toggle_online_status( window, NULL );

	screem_window_change_view( window, details->editor );

	bonobo_ui_component_thaw( details->shell_ui, NULL );
}

Bonobo_UIContainer screem_window_get_ui_container( ScreemWindow *window )
{
	g_return_val_if_fail( SCREEM_IS_WINDOW( window ),
			      CORBA_OBJECT_NIL );

	return bonobo_object_corba_objref(BONOBO_OBJECT(window->details->ui_container));
}

BonoboUIComponent *screem_window_get_shell_ui( ScreemWindow *window )
{
	g_return_val_if_fail( SCREEM_IS_WINDOW( window ), NULL );

	return window->details->shell_ui;
}

void screem_window_change_view( ScreemWindow *window,
				GtkWidget *view )
{
	gboolean activate = FALSE;
	ScreemWindowDetails *details;

	details = window->details;

	if( view == details->editor &&
	    details->current != details->editor ) {
		/* switch to it */
		gtk_widget_hide( details->current );
		details->current = details->editor;
		activate = TRUE;
	} else if( view == details->preview &&
		   details->current != details->preview ) {
		/* switch to it */
		gtk_widget_hide( details->current );
		details->current = details->preview;
		activate = TRUE;
	} else if( view == details->link_view &&
		   details->current != details->link_view ) {
		/* switch to it */
		gtk_widget_hide( details->current );
		details->current = details->link_view;
		activate = TRUE;
	} else if( view == details->tree_view &&
		   details->current != details->tree_view ) {
		/* switch to it */
		gtk_widget_hide( details->current );
		details->current = details->tree_view;
		activate = TRUE;
	} else {
		activate = FALSE;
	}
		
	if( activate ) {
		gtk_widget_show( details->current );
		screem_window_display_view( window );
	} 
}

void screem_window_add_site( ScreemWindow *window, ScreemSite *site )
{
	const gchar *name;
	gchar *path;
	BonoboUIComponent* ui;
	GClosure *closure;
	const gchar *pathname;
	SwitchSiteData *sd;
	gchar *verb_name;
	gchar *cmd;
	RBSidebarButton *button;
	
	name = screem_site_get_name( site );
	
	button = rb_sidebar_button_new( name, name );
	g_object_set_data( G_OBJECT( button ), "site", site );

	rb_sidebar_button_set( button, "Screem_Site", name, TRUE );
	rb_sidebar_append( RB_SIDEBAR( window->details->sitebar ),
			   button );
	gtk_widget_show( GTK_WIDGET( button ) );

	sd = g_new0( SwitchSiteData, 1 );
	sd->site = site;
	sd->window = window;
	g_signal_connect( G_OBJECT( button ), "clicked",
			  G_CALLBACK( screem_window_site_switch ), sd );

	window->details->sitebargroup = 
		g_slist_prepend( window->details->sitebargroup, button );

	verb_name = g_strconcat( "SwitchSite_", name, NULL );
	path = g_strconcat( "/menu/Site/SwitchTo/",
			    verb_name,
			    NULL );
	
	ui = screem_window_get_shell_ui( window );
 	
	bonobo_ui_component_freeze( ui, NULL );

	pathname = screem_site_get_pathname( site );
	if( ! pathname ) {
		pathname = _( "Individual Files" );
	}
	
	cmd = g_strdup_printf ("<cmd pixtype=\"stock\" pixname=\"Screem_Site\" name=\"%s\" /> ", verb_name);
	bonobo_ui_component_set_translate( ui, "/commands/", cmd, 
					   NULL );
	
	closure = g_cclosure_new( G_CALLBACK( screem_window_switch_site ),
				  sd, 
				  screem_window_switch_site_destroy );
	bonobo_ui_component_add_verb_full( ui, verb_name, closure ); 
	
	if( bonobo_ui_component_path_exists( ui, path, NULL ) ) {
		bonobo_ui_component_set_prop( ui, path,
					      "label", name, NULL );
		bonobo_ui_component_set_prop( ui, path,
					      "tip", pathname, NULL );
	} else {
		gchar *xml;
		
		xml = g_strconcat( "<menuitem name=\"", verb_name,
				   "\" verb=\"", verb_name, 
				   "\" _label=\"", name,
				   "\" _tip=\"", pathname,
				   "\" />", NULL );
		
		bonobo_ui_component_set_translate( ui, 
						   "/menu/Site/SwitchTo",
						   xml, NULL );
		g_free( xml );
	}
	bonobo_ui_component_thaw( ui, NULL );

	g_free( verb_name );
	g_free( path );
}


void screem_window_set_current( ScreemWindow *window, ScreemSite *site )
{
        const gchar *site_name;
	RBSidebarButton *button;
	GSList *sites;

	GList *docs;

	GtkTreeViewColumn *column;
	GtkTreeModel *model;
	GtkTreePath *path;
	GtkWidget *tree;

        g_return_if_fail( SCREEM_IS_WINDOW( window ) );
        g_return_if_fail( SCREEM_IS_SITE( site ) );

	if( site == window->details->current_site ) {
		return;
	}

	site_name = screem_site_get_name( site );

     	for( sites = window->details->sitebargroup; sites;
	     sites = sites->next ) {
		button = RB_SIDEBAR_BUTTON( sites->data );
		if( ! strcmp( site_name, button->unique_id ) ) {
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( button ),
						      TRUE );
			break;
		}
	}

	model = screem_site_get_model( site );
	tree = GTK_WIDGET( g_object_get_data( G_OBJECT( window ), "siteview" ) );
	g_object_ref( G_OBJECT( model ) );
	gtk_tree_view_set_model( GTK_TREE_VIEW( tree ), model );

        if( window->details->current_site ) {
		/* store documents list */
		screem_site_set_documents( window->details->current_site, 
					   window->details->name,
					   window->details->documents );
		docs = window->details->documents;
		window->details->documents = NULL;
		for( ; docs; docs = docs->next ) {
			/* always remove 0, as the notebook will
			   switch to the next page itself */
			gtk_notebook_remove_page( GTK_NOTEBOOK( window->details->notebook ), 0 );
		}
        }

        window->details->current_site = site;

	column = gtk_tree_view_get_column( GTK_TREE_VIEW( tree ), 0 );
	gtk_tree_view_column_set_title( column, screem_site_get_name( site ) );

	g_object_set_data( G_OBJECT( tree ), "site", site );

	column = gtk_tree_view_get_column( GTK_TREE_VIEW( tree ), 1 );

	gtk_tree_view_column_set_visible( column, 
					  ! screem_site_get_fake_flag(site) );
	
	/* expand root node */
	path = gtk_tree_path_new_first();
	gtk_tree_view_expand_row( GTK_TREE_VIEW( tree ),
				  path, FALSE );
	gtk_tree_path_free( path );


	gtk_notebook_set_current_page(GTK_NOTEBOOK(window->details->file_view),
				      0);

	/* now loop through details->documents and insert each page */
	docs = screem_site_get_documents( site, window->details->name );

	if( docs ) {
		for( window->details->documents = NULL; docs; 
		     docs = docs->next )
			screem_window_set_document( window, 
						    SCREEM_PAGE(docs->data) );
	} else {
		screem_window_set_document( window, NULL );
	}


	{
		/* enable || disable ui elements */
		gboolean enable = ( ! screem_site_get_fake_flag( site ) );
		BonoboUIComponent *ui;
		static const gchar* elements[] = {
			"/commands/Close Site",

			"/commands/SiteSettings",
			"/commands/UpdateSite",
			"/commands/ImportSite",
			"/commands/CommitSite",
			"/commands/TaskList",
			NULL
		};
		gint i;

		ui = screem_window_get_shell_ui( SCREEM_WINDOW( window ) );
		bonobo_ui_component_freeze( ui, NULL );

		for( i = 0; elements[ i ]; ++ i ) {
			bonobo_ui_component_set_prop( ui,
						      elements[ i ],
						      "sensitive",
						      ( enable ? "1" : "0" ),
						      NULL );
		}
		bonobo_ui_component_thaw( ui, NULL );
	}
}

ScreemSite *screem_window_get_current( ScreemWindow *window )
{
	return window->details->current_site;
}

void screem_window_set_document( ScreemWindow *window, ScreemPage *page )
{
        const gchar *pathname;
        ScreemWindowDetails *details;
	ScreemSite *site;

	gboolean loaded = TRUE;

        g_return_if_fail( SCREEM_IS_WINDOW( window ) );

        details = window->details;

	if( ! details->current ) {
		/* no current view, don't display, probably
		   destroying the window */
		return;
	}

	if( page != details->current_document &&
	    details->current_document ) {
		gint tab;
		GtkWidget *child;

		tab = g_list_index( details->documents, 
				    details->current_document );
		if( tab > -1 ) {
			gint pos;

			child = gtk_notebook_get_nth_page( GTK_NOTEBOOK( details->notebook ),
							   tab );
			child = gtk_notebook_get_tab_label( GTK_NOTEBOOK( details->notebook ), child );
			pos = screem_editor_get_pos( SCREEM_EDITOR( details->editor ) );

			g_object_set_data( G_OBJECT( child ), "cursor_pos", 
					   GINT_TO_POINTER( pos ) );
		}

		/* disconnect doctype_changed */
		g_signal_handlers_disconnect_matched(G_OBJECT( details->current_document ),
						     G_SIGNAL_MATCH_FUNC |
						     G_SIGNAL_MATCH_DATA,
						     0, 0, NULL, 
						     screem_window_set_doctype,
						     window );
	}

	site = screem_window_get_current( window );
        details->current_document = page;

	if( page && screem_site_get_fake_flag( site ) ) {
		pathname = screem_page_get_pathname( page );
		if( pathname ) {
			gedit_recent_add( details->recent_pages, pathname );
		}
	}

     	if( page && ! screem_window_set_document_tab( window, page ) ) {
		/* new tab */
                GtkWidget *label;
                GtkWidget *hack;
                GtkWidget *event;
                GtkTooltips *tip;

                details->documents = g_list_append( details->documents, page );

		pathname = screem_page_get_pathname( page );
		
		if( ! pathname || screem_page_load( page ) ) {
			gboolean changed;

			label = gtk_label_new( "" );

			changed = screem_page_get_changed( page );
			screem_window_set_page_label( page, changed, label );
			
			g_signal_connect( G_OBJECT( page ),
					  "changed",
					  G_CALLBACK( screem_window_set_page_label ),
					  label );

			hack = gtk_hbox_new(0,FALSE);
			gtk_widget_show( hack );
			gtk_widget_show( label );
			event = gtk_event_box_new();
		
			gtk_container_add( GTK_CONTAINER( event ), label );
			g_object_set_data( G_OBJECT( event ), "document", 
					   page );
			g_signal_connect( G_OBJECT( event ), 
					  "button_press_event",
					  G_CALLBACK( screem_window_tab_clicked ),
					  window );
			gtk_notebook_append_page( GTK_NOTEBOOK( details->notebook ),
						  hack, event );
			gtk_widget_set_size_request( hack, 0, 0 );
			/* set tooltip to full pathname */
			tip = gtk_tooltips_new();
			if( ! pathname ) {
				pathname = _( "Unsaved Document" );
			}
			gtk_tooltips_set_tip( tip, event, pathname, 
					      NULL );
			/* switch to the page */
			
			gtk_notebook_set_current_page( GTK_NOTEBOOK(details->notebook),
						       g_list_length( details->documents ) - 1 );
			screem_page_set_open( page, TRUE );
		} else {
			/* failed to open */
			GtkWidget *alert;
			
			pathname = screem_page_get_pathname( page );

			alert = gtk_message_dialog_new( GTK_WINDOW( window ),
							GTK_DIALOG_MODAL |
							GTK_DIALOG_DESTROY_WITH_PARENT,
							GTK_MESSAGE_ERROR,
							GTK_BUTTONS_OK,
							_("Failed to load %s"),
							pathname );
			gtk_window_set_transient_for( GTK_WINDOW( alert ),
						      GTK_WINDOW( window ) );
			gtk_dialog_run( GTK_DIALOG( alert ) );
			gtk_widget_destroy( alert );
			loaded = FALSE;
		}
        }
	if( loaded ) {
		screem_window_display_view( window );

		if( page ) {
			const gchar *pathname;
			const gchar *doctype;
			gchar *title;
			GtkWidget *child;
			gint tab;
			gint pos;

			tab = gtk_notebook_get_current_page( GTK_NOTEBOOK( details->notebook ) );

			child = gtk_notebook_get_nth_page( GTK_NOTEBOOK( details->notebook ),
							   tab );
			child = gtk_notebook_get_tab_label( GTK_NOTEBOOK( details->notebook ), child );
			pos = GPOINTER_TO_INT( g_object_get_data( G_OBJECT( child ),
								  "cursor_pos" ) );
	
			/*	screem_editor_set_pos( SCREEM_EDITOR( details->editor ),
				pos );*/

			pathname = screem_page_get_pathname( page );
			if( ! pathname ) {
				pathname = _( "Untitled" );
			}

			title = g_strdup_printf( "%s - Screem",
						 pathname );
			gtk_window_set_title( GTK_WINDOW( window ),
					      title );
			g_free( title );

			doctype = screem_page_get_doctype( page );

			screem_window_set_doctype( page, doctype, window );
			g_signal_connect(G_OBJECT( page ), "doctype_changed",
					 G_CALLBACK(screem_window_set_doctype),
					 window );
		} else {
			gtk_window_set_title( GTK_WINDOW( window ),
					      "Screem" );
		}
	}

	{
		/* enable || disable ui elements */
		gboolean enable = ( page && loaded );
		BonoboUIComponent *ui;
		static const gchar* elements[] = {
			"/commands/Save Document",
			"/commands/Save Document As",
			"/commands/Close Document",

			"/commands/Print",
			"/commands/PrintPreview",

			"/commands/Undo", "/commands/Redo",
			"/commands/Cut", "/commands/Copy",
			"/commands/Paste", "/commands/PasteEncoded",
			"/commands/Clear", "/commands/SelectAll",
			"/commands/SelectContext",
			"/commands/Search","/commands/Bookmarks",
			"/commands/SpellCheck",
			"/commands/Browser0", "/commands/Browser1",
			"/commands/Browser2", "/commands/Browser3",

			"/commands/Bold", "/commands/Italic",
			"/commands/Underline", "/commands/Para",
			"/commands/Bullet", "/commands/Numbered",
			"/commands/Unindent", "/commands/Indent",
			"/commands/AlignLeft", "/commands/AlignCentre",
			"/commands/AlignRight", "/commands/Justify",
			"/commands/Font", "/commands/Pre",
			"/commands/Subscript", "/commands/Superscript",
			"/commands/Table", "/commands/TBody",
			"/commands/Tr", "/commands/Th", "/commands/Td",
			"/commands/Caption", "/commands/Summary",
			"/commands/Textarea", "/commands/Input",
			"/commands/Fileinput", "/commands/Button",
			"/commands/Checkbutton", "/commands/Radiobutton",
			"/commands/Optionmenu",
			
			"/commands/CommitPage", "/commands/UpdatePage",

			"/menu/Insert",
			NULL
		};
		gint i;

		ui = screem_window_get_shell_ui( SCREEM_WINDOW( window ) );
		bonobo_ui_component_freeze( ui, NULL );

		for( i = 0; elements[ i ]; ++ i ) {
			bonobo_ui_component_set_prop( ui,
						      elements[ i ],
						      "sensitive",
						      ( enable ? "1" : "0" ),
						      NULL );
		}
		bonobo_ui_component_thaw( ui, NULL );
	}
}

ScreemPage* screem_window_get_document( ScreemWindow *window )
{
	return window->details->current_document;
}

gboolean screem_window_set_document_tab( ScreemWindow *window, ScreemPage *page )
{
	ScreemWindowDetails *details;
	GnomeVFSURI *uri;

	details = window->details;

	if( g_list_find( details->documents, page ) ) {
		GtkWidget *event;
		GList *children;
		GtkWidget *label;
		const gchar *filename;
		GtkWidget *notebook_page;
		GtkTooltipsData *ttip;
		const gchar *pathname;
		gint index;

		/* its open so we can do something with it */
                index = g_list_index( details->documents, page );

		if( gtk_notebook_get_current_page( GTK_NOTEBOOK( details->notebook ) ) != index ) {
			gtk_notebook_set_current_page( GTK_NOTEBOOK( details->notebook ), index );
			return TRUE;
		}

		/* but we want to ensure the tooltip has
		   the correct pathname */
		notebook_page = 
			gtk_notebook_get_nth_page( GTK_NOTEBOOK( details->notebook ), index );
		event = gtk_notebook_get_tab_label( GTK_NOTEBOOK( details->notebook ),
						    notebook_page );
		ttip = gtk_tooltips_data_get( event );
		pathname = screem_page_get_pathname( page );
		gtk_tooltips_set_tip( ttip->tooltips, event,
				      pathname, "" );
		/* get label from event */
		children = gtk_container_children( GTK_CONTAINER( event ) );
		label = children->data;
		uri = gnome_vfs_uri_new( pathname );
		filename = gnome_vfs_uri_extract_short_name( uri );
		gnome_vfs_uri_unref( uri );
		gtk_label_set_text( GTK_LABEL( label ), filename );

		return TRUE;
        }

	return FALSE;
}

void screem_window_remove_document( ScreemWindow *window, ScreemPage *page )
{
	ScreemWindowDetails *details;
	GtkWidget *notebook;
	gint index;
       
	details = window->details;
	notebook = details->notebook;

	if( g_list_find( details->documents, page ) ) {
		index = g_list_index( details->documents, page );
		
		details->documents = g_list_remove( details->documents, page );
		if( page == details->current_document ) {
			details->current_document = NULL;
		}
		gtk_notebook_remove_page( GTK_NOTEBOOK( notebook ), index );
	}
}

void screem_window_show_message( ScreemWindow *window, const gchar *message )
{
	GladeXML *xml;
	GtkWidget *widget;

	xml = glade_get_widget_tree( window->details->messages );
	widget = glade_xml_get_widget( xml, "messages_text" );

	screem_window_real_show_message( window, widget, message );
}

void screem_window_show_error( ScreemWindow *window, const gchar *error )
{
	GladeXML *xml;
	GtkWidget *widget;

	xml = glade_get_widget_tree( window->details->messages );
	widget = glade_xml_get_widget( xml, "error_text" );

	screem_window_real_show_message( window, widget, error );
}


void screem_window_clear_messages( ScreemWindow *window)
{
	GladeXML *xml;
	GtkWidget *widget;

	xml = glade_get_widget_tree( window->details->messages );
	widget = glade_xml_get_widget( xml, "message_text" );

	screem_window_real_clear_messages( widget );
}

void screem_window_clear_errors( ScreemWindow *window )
{
	GladeXML *xml;
	GtkWidget *widget;

	xml = glade_get_widget_tree( window->details->messages );
	widget = glade_xml_get_widget( xml, "error_text" );

	screem_window_real_clear_messages( widget );
}

void screem_window_print( ScreemWindow *window, gboolean preview )
{
	ScreemWindowDetails *details;

	details = window->details;

	screem_view_print( SCREEM_VIEW( details->current ),
			   preview );
}


void screem_window_add_side_tab( ScreemWindow *window,
				 GtkWidget *label,
				 GtkWidget *icon,
				 GtkWidget *content )
{
	GtkWidget *tbox;

	tbox = gtk_hbox_new(0,FALSE);
	gtk_box_pack_start( GTK_BOX( tbox ), icon, 
			    TRUE, TRUE, 0 );
	/*gtk_box_pack_start( GTK_BOX( tbox ), label, 
	  TRUE, TRUE, 0 );*/
	gtk_widget_show_all( tbox );
		
	gtk_widget_show_all( content );

	gtk_notebook_append_page( GTK_NOTEBOOK( window->details->file_view ),
				  content, tbox );
}


ScreemDTD *screem_window_get_current_dtd( ScreemWindow *window )
{
	g_return_val_if_fail( SCREEM_IS_WINDOW( window ), NULL );

	return window->details->current_dtd;
}

void screem_window_display_view( ScreemWindow *window )
{
	ScreemSite *site;
	ScreemPage *page;
	ScreemWindowDetails *details;

	details = window->details;
	site = screem_window_get_current( window );
	page = details->current_document;

	if( ! site ) {
		return;
	}

	if( details->current ) {
		g_object_set( G_OBJECT( details->current ), "online",
			      ! window->details->offline, NULL );
		
		screem_view_display( SCREEM_VIEW( details->current ) );

		if( details->current != details->tree_view ) {
			screem_view_display( SCREEM_VIEW( details->tree_view ) );
		}
	}
}


void screem_window_toggle_online_status( ScreemWindow *window, gpointer data )
{
	ScreemApplication *application;
	GConfClient *client;
	gboolean offline;
	BonoboUIComponent *ui;

	application = window->application;
	client = application->client;

	ui = screem_window_get_shell_ui( window );

	if( data ) {
		/* toggle actual setting */
		application->offline = ! application->offline;
	}

	offline = application->offline;

	if( window->details->offline != offline ) {
		gconf_client_set_bool( client,
				       "/apps/screem/general/work_offline",
				       offline,
				       NULL );
		window->details->offline = offline;
	}
	bonobo_ui_component_set_prop( ui,
				      "/commands/Offline",
				      "hidden", offline ? "1" : "0", NULL );
	bonobo_ui_component_set_prop( ui,
				      "/commands/Online",
				      "hidden", offline ? "0" : "1", NULL );

	gtk_image_set_from_stock( GTK_IMAGE( window->details->online_status ),
				  offline ? "Screem_Offline" : "Screem_Online",
				  GTK_ICON_SIZE_BUTTON );
}

static void screem_window_real_show_message( ScreemWindow *window,
					     GtkWidget *view, 
					     const gchar *message )
{
	ScreemApplication *application;
	GtkTextBuffer *buffer;
	GtkTextIter it;
	GtkTextIter eit;
	gint buffersize;
	BonoboUIComponent *uic;
	GtkWidget *widget;
	GtkNotebook *notebook;
	gint i;

	g_return_if_fail( GTK_IS_TEXT_VIEW( view ) );

	application = SCREEM_APPLICATION( window->application );
	buffersize = gconf_client_get_int( application->client,
					   "/apps/screem/messagelines",
					   NULL );

	/* put in status bar, or message box */
	if( ! strchr( message, '\n' ) ) {
		uic = screem_window_get_shell_ui( window );
		bonobo_ui_component_set_status( uic, message, NULL );
	} else {
		buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
		
		/* first trim the buffer if needed*/
		gtk_text_buffer_get_end_iter( buffer, &eit );
		if( gtk_text_iter_get_line( &eit ) > buffersize ) {
			gtk_text_buffer_get_iter_at_line( buffer, &it, 0 );
			gtk_text_buffer_get_iter_at_line( buffer, &eit, 1 );	
			gtk_text_buffer_delete( buffer, &it, &eit );
		}
		
		/* insert the message */
		gtk_text_buffer_get_end_iter( buffer, &eit );
		gtk_text_buffer_insert( buffer, &eit, message, 
					strlen( message ) );
		gtk_text_buffer_get_end_iter( buffer, &eit );
		gtk_text_buffer_insert( buffer, &eit, "\n", strlen( "\n" ) );
	}

	/* now find out which page in window->details->messages
	   that view occurs, and switch to it */
	notebook = GTK_NOTEBOOK( window->details->messages );
	for( i = 0; (widget = gtk_notebook_get_nth_page(notebook, i )); ++i ) {
		if( widget == view || 
		    widget == gtk_widget_get_parent( view ) ) {
			gtk_notebook_set_current_page( notebook, i );
			break;
		}
	}

	/* set timeout to clear status bar */
	g_timeout_add( 2000, (GSourceFunc)screem_window_clear_status, window );
}

static void screem_window_real_clear_messages( GtkWidget *view )
{
	GtkTextBuffer *buffer;
	GtkTextIter it;
	GtkTextIter eit;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));

	gtk_text_buffer_get_start_iter( buffer, &it );
	gtk_text_buffer_get_end_iter( buffer, &eit );
	gtk_text_buffer_delete( buffer, &it, &eit );

	buffer = gtk_text_view_get_buffer( GTK_TEXT_VIEW( view ) );
}

static gboolean screem_window_clear_status( ScreemWindow *window )
{
	BonoboUIComponent *uic;

	uic = screem_window_get_shell_ui( window );
	bonobo_ui_component_set_status( uic, "", NULL );

	return FALSE;
}

static gboolean screem_window_file_view_click( GtkTreeView *tree_view,
					       GtkTreePath *path,
					       GtkTreeViewColumn *column,
					       gpointer data )
{
	ScreemWindow *window;
	ScreemSite *site;

	GtkTreeSelection *sel;

	window = SCREEM_WINDOW( data );
	site = screem_window_get_current( window );

	sel = gtk_tree_view_get_selection( GTK_TREE_VIEW( tree_view ) );

	if( sel ) {
		ScreemPage *page;

		const gchar *current;

		current = g_object_get_data( G_OBJECT( tree_view ), 
					     "file_activated" );
		if( current && ! screem_uri_is_dir( current ) ) {
			page = screem_site_locate_page( site, current );
			if( page ) {
				screem_window_set_document( window, page );
			} else if( screem_site_get_fake_flag( site ) ) {
				screem_page_open_with_filename( site, window,
								current );
			}
			g_object_set_data( G_OBJECT( tree_view ),
					   "file_activated", NULL );
		}
		
	} else {
		/* hmm, no iterator */
		
	}

	return (sel != NULL);
}

static gboolean screem_window_file_view_press( GtkWidget *widget,
					       GdkEventButton *event,
					       ScreemWindow *window )
{
	GtkTreePath *path;
	GtkTreeViewColumn *column;
	gint cx;
	gint cy;

	GtkTreeModel *model;
	GtkTreeIter it;

	GValue value = {0,};
	ScreemSiteViewNodeInfo *info;

	ScreemSite *site;

	if( event && event->button != 3 )
		return FALSE;

	site = screem_window_get_current( window );

	/* right mouse clicked */
	gtk_tree_view_get_path_at_pos( GTK_TREE_VIEW( widget ), 
				       event->x, event->y, 
				       &path, &column, &cx, &cy );

	if( path ) {
		GtkTreeSelection *selection;

		if( event )
			g_signal_stop_emission_by_name( G_OBJECT( widget ),
							"button_press_event" );
		
		model = gtk_tree_view_get_model( GTK_TREE_VIEW( widget ) );
		gtk_tree_model_get_iter( GTK_TREE_MODEL( model ), &it, path );
		
		selection = gtk_tree_view_get_selection( GTK_TREE_VIEW( widget ) );
		gtk_tree_selection_select_iter( GTK_TREE_SELECTION( selection ),
						&it );
		
		if( ! screem_site_get_fake_flag( site ) ) {
			gtk_tree_model_get_value( model, &it, FILE_BROWSER_DATA_COL, &value );
			info = g_value_get_pointer( &value );
		} else {
			info = NULL;
		}
		if( info ) {
			ScreemPage *page;
			GString *pathname;
			ScreemSite *site;
			GtkWidget *menu;
			BonoboUIComponent *ui;

			const gchar *exclude_path;
			const gchar *ignore_path;
			const gchar *ascii_path;
			gboolean status;
			gchar *state;
			gboolean exists;

			exclude_path = "/commands/ExcludeFlag";
			ignore_path = "/commands/IgnoreFlag";
			ascii_path = "/commands/ASCIIFlag";

			pathname = g_string_new( info->fullname );
			
			site = screem_window_get_current( window );
			
			page = screem_site_locate_page( site, pathname->str );
			
			/* set upload status toggle buttons */
			ui = screem_window_get_shell_ui( window );
			bonobo_ui_component_freeze( ui, NULL );

			status = screem_site_is_excluded( site, info->fullname );
			bonobo_ui_component_set_prop( ui, exclude_path,
						      "state", 
						      status ? "1" : "0",
						      NULL );
			status = screem_site_is_ignored( site, info->fullname );
			bonobo_ui_component_set_prop( ui, ignore_path,
						      "state", 
						      status ? "1" : "0",
						      NULL );
			status = screem_site_is_ascii( site, info->fullname );
			bonobo_ui_component_set_prop( ui, ascii_path,
						      "state", 
						      status ? "1" : "0",
						      NULL );
			bonobo_ui_component_thaw( ui, NULL );
			menu = gtk_menu_new();
			bonobo_window_add_popup( BONOBO_WINDOW( window ), 
						 GTK_MENU( menu ),
						 "/popups/siteviewmenu" );
			gnome_popup_menu_do_popup_modal( menu, 0, 0, event,
							 0, 0 );

			exists = uri_exists( pathname->str, NULL );

			state = bonobo_ui_component_get_prop( ui,
							      exclude_path,
							      "state",
							      NULL );
			if( exists && *state == '1' ) {
				screem_site_add_exclude( site, pathname->str );
			} else {
				screem_site_remove_exclude( site, 
							    pathname->str );
			}
			g_free( state );
			state = bonobo_ui_component_get_prop( ui,
							      ignore_path,
							      "state",
							      NULL );
			if( exists && *state == '1' ) {
				screem_site_add_ignore( site, pathname->str );
			} else {
				screem_site_remove_ignore( site, 
							   pathname->str );
			}
			g_free( state );
			state = bonobo_ui_component_get_prop( ui,
							      ascii_path,
							      "state",
							      NULL );
			if( exists && *state == '1' ) {
				screem_site_add_ascii( site, pathname->str );
			} else {
				screem_site_remove_ascii( site, 
							  pathname->str );
			}
			g_free( state );

			g_string_free( pathname, TRUE );

			gtk_widget_destroy( menu );

		}
		g_value_unset( &value );

		gtk_tree_path_free( path );
	}
	return ( path != NULL );
}

static void screem_window_tab_clicked( GtkWidget *label, GdkEventButton *event,
                                       ScreemWindow *window )
{
        ScreemPage *page;
        ScreemWindowDetails *details;
        gint num;

        details = window->details;

        page = SCREEM_PAGE( g_object_get_data( G_OBJECT( label ),
					       "document" ) );

        num = g_list_index( details->documents, page );

        if( ( ! event ) || event->button != 3 ) {
                gtk_notebook_set_current_page( GTK_NOTEBOOK( details->notebook ), 
					       num );        
	} else {
                /* right click on page tab, show popup */
		GtkWidget *menu;
		GtkWidget *submenu;
		GtkWidget *menuitem;
		GList *docs;
		
		menu = gtk_menu_new();

		bonobo_window_add_popup( BONOBO_WINDOW( window ), 
					GTK_MENU( menu ),
					 "/popups/tabmenu" );
		g_object_set_data( G_OBJECT( window ), "tab_popup", page );
		
		/* add open pages menu */
		menuitem = gtk_menu_item_new_with_label( _( "Opened Pages" ) );
		gtk_widget_show( menuitem );
		gtk_menu_shell_append( GTK_MENU_SHELL( menu ), menuitem );
		submenu = gtk_menu_new();
		gtk_menu_item_set_submenu( GTK_MENU_ITEM( menuitem ), submenu );
		
		for( docs = details->documents; docs; docs = docs->next ) {
			const gchar *pathname;
			gchar *base;
			page = SCREEM_PAGE( docs->data );
			pathname = screem_page_get_pathname( page );
			
			if( ! pathname ) {
				pathname = _( "Untitled" );
			}
			base = g_path_get_basename( pathname );
			menuitem = gtk_menu_item_new_with_label( base );
			gtk_widget_show( menuitem );
			g_object_set_data( G_OBJECT( menuitem ), "document",
					   page );
			g_signal_connect( G_OBJECT( menuitem ), "activate",
					G_CALLBACK( screem_window_tab_popup_switch ),
					window );
			gtk_menu_shell_append( GTK_MENU_SHELL( submenu ),
						menuitem );
			g_free( base );
		}
		
		gnome_popup_menu_do_popup_modal( menu, 0, 0, event, 0, label );
		g_object_set_data( G_OBJECT( window ), "tab_popup", NULL );
		gtk_widget_destroy( menu );
        }
}

static void screem_window_tab_popup_switch( GtkWidget *widget, gpointer data )
{
	screem_window_tab_clicked( widget, NULL, data );	
}

static void screem_window_notebook_change( GtkNotebook *notebook, 
					   GtkNotebookPage *p,
                                           guint num )
{
        ScreemWindow *window;
	ScreemWindowDetails *details;
        ScreemPage *page;
        GList *list;
	

        window = SCREEM_WINDOW( g_object_get_data( G_OBJECT( notebook ),
						   "window" ) );
	details = window->details;

	/* must be destroying, don't bother doing anything */
	if( ! details->current ) {
		return;
	}

	if( details->documents ) {
		gint len;

		len = g_list_length( details->documents ) - 1;

		if( num > len ) {
			num = len;
		}

		list = g_list_nth( details->documents, num );

		if( list ) {
			g_assert( SCREEM_IS_PAGE( list->data ) );

			page = SCREEM_PAGE( list->data );
			/* if page == screem_window_get_document then don't try
			   to set it again */
			if( page != screem_window_get_document( window ) ) {
				screem_window_set_document( window, page );
			}
		} else {
			screem_window_set_document( window, NULL );
		}
	} else {
		screem_window_set_document( window, NULL );
	}
}


static void screem_window_highlight_spelling( ScreemSpell *spell,
					      gint start, gint end,
					      ScreemWindow *window )
{
	ScreemEditor *editor;

	editor = SCREEM_EDITOR( window->details->editor );

	screem_editor_set_pos( editor, start );
	screem_editor_select_region( editor, start, end - start );
}


static void screem_window_shortcut_notify( GConfClient *client,
					   guint cnxn_id,
					   GConfEntry *entry,
					   gpointer data )
{
	ScreemWindow *window;

	window = SCREEM_WINDOW( data );

	if( entry->value->type == GCONF_VALUE_BOOL ) {
		gboolean state = gconf_value_get_bool( entry->value );

		if( state && 
		    ! GTK_WIDGET_VISIBLE( window->details->barbox ) ) {
			gtk_widget_show( window->details->barbox );
		} else if( GTK_WIDGET_VISIBLE( window->details->barbox ) ) {
			gtk_widget_hide( window->details->barbox );
		}
	}
}

static void screem_window_set_page_label( ScreemPage *page,
					  gboolean status,
					  gpointer data )
{
	GtkWidget *label;
	const gchar *pathname;
	GnomeVFSURI *uri;
	gchar *filename;
	gchar *txt;
	
	if( screem_page_get_changed( page ) != status ||
	    ! GTK_WIDGET_REALIZED( GTK_WIDGET( data ) ) ) {
		label = GTK_WIDGET( data );
		
		pathname = screem_page_get_pathname( page );
		if( pathname && ( strlen( pathname ) > 0 ) ) {
			uri = gnome_vfs_uri_new( pathname );
			filename = gnome_vfs_uri_extract_short_name( uri );
			gnome_vfs_uri_unref( uri );
		} else {
			filename = g_strdup( _( "Untitled" ) );
		}
		
		if( status ) {
			txt = g_strconcat( "* ", filename, NULL );
		} else {
			txt = g_strdup( filename );
		}
		gtk_label_set_text( GTK_LABEL( label ), txt );
		
		g_free( filename );
		g_free( txt );
	}
}

static void screem_window_set_doctype( ScreemPage *page,
				       const gchar *doctype,
				       gpointer data )
{
	ScreemApplication *application;
	ScreemWindow *window;
	ScreemDTD *dtd;
     
	window = SCREEM_WINDOW( data );
	application = window->application;

	if( doctype ) {
		dtd = screem_application_add_dtd( application, doctype, NULL );
	} else {
		dtd = NULL;
	}

	if( ! dtd ) {
		/* use default */
		gchar *ddtd;

		ddtd=gconf_client_get_string(application->client,
					     "/apps/screem/editor/default_dtd",
					     NULL );
		if( ! ddtd ) {
			ddtd = g_strdup( "-//W3C//DTD HTML 4.01 Transitional//EN" );
		}

		dtd = screem_application_add_dtd( application, ddtd, NULL );
	
		g_free( ddtd );
	}

	if( window->details->current_dtd != dtd ) {
		window->details->current_dtd = dtd;
		
		screem_tag_tree_add_dtd( SCREEM_TAG_TREE( window->details->tagtree ),
					 dtd );
	}
}

static void screem_window_select_context_callback( ScreemTreeView *view, 
						   guint start,
						   guint end, gpointer data )
{
	ScreemWindow *window;
	ScreemEditor *editor;

	window = SCREEM_WINDOW( data );
	editor = SCREEM_EDITOR( window->details->editor );

	if( GTK_WIDGET( editor ) == window->details->current ) {
		screem_editor_select_region( editor, start,
					     end - start );
	}
}

static void screem_window_request_context_callback( ScreemTreeView *view, 
						    guint pos,
						    gchar **context_name,
						    guint *depth,
						    guint *start,
						    guint *end,
						    gboolean build,
						    gpointer data )
{
	ScreemWindow *window;
	ScreemTreeView *tree;
	gchar *tag;

	window = SCREEM_WINDOW( data );
	tree = SCREEM_TREE_VIEW( window->details->tree_view );

	tag = screem_tree_view_query_element_context( tree, pos,
						      FALSE, build,
						      depth, start, end );
	if( context_name ) {
		*context_name = tag;
	} else {
		g_free( tag );
	}
}

static void screem_window_view_change( RBSidebarButton *button, 
				       ScreemWindow *window )
{
	GtkWidget *view;

	view = NULL;

	if( ! strcmp( "editor", button->unique_id ) ) {
		view = window->details->editor;
	}
	if( ! strcmp( "preview", button->unique_id ) ) {
		view = window->details->preview;
	}
	if( ! strcmp( "linkview", button->unique_id ) ) {
		view = window->details->link_view;
	}
	if( ! strcmp( "treeview", button->unique_id ) ) {
		view = window->details->tree_view;
	}

	g_assert( view );

	screem_window_change_view( window, view );
}

static void
screem_window_switch_site_destroy( gpointer data, GClosure *closure )
{
	SwitchSiteData *sd = ( SwitchSiteData * ) data;

	g_return_if_fail( sd );

	g_free( sd );
}

static void
screem_window_switch_site( BonoboUIComponent *uic, 
			   gpointer data, const char *cname )
{
	SwitchSiteData *sd = ( SwitchSiteData * ) data;
	ScreemWindow *window;
	ScreemSite *site;

	g_return_if_fail( sd );

	window = sd->window;
	site = sd->site;

	g_assert( window );
	g_assert( site );

	screem_window_set_current( window, sd->site );
}

static void screem_window_site_switch( RBSidebarButton *button,
				       SwitchSiteData *sd )
{
	screem_window_set_current( sd->window, sd->site );
}

static void screem_window_drop( GtkWidget *widget, GdkDragContext *context,
				gint x, gint y, 
				GtkSelectionData *selectionData,
				guint info, guint time, ScreemWindow *window )
{
	ScreemApplication *app;
	gboolean handled;
	GList *uris;
	GList *tmp;
	gchar *text;

	handled = FALSE;

	app = window->application;

	if( info == TARGET_MOZ ) {
		glong read;
		glong written;

		text = g_utf16_to_utf8( selectionData->data,
					selectionData->length,
					&read, &written,
					NULL );
	} else {
		text = g_strdup( selectionData->data );
	}

	switch( info ) {
	case TARGET_MOZ:
	case TARGET_URL:
	case TARGET_URI_LIST:
		uris = gnome_vfs_uri_list_parse( text );
		for( tmp = uris; tmp; tmp = tmp->next ) {
			GnomeVFSURI *uri;
			gchar *url;

			uri = (GnomeVFSURI*)tmp->data;
			url = gnome_vfs_uri_to_string( uri, 0 );

			if( ( ! window->details->offline ) ||
			    gnome_vfs_uri_is_local( uri ) ) {
				ScreemSite *site;
				ScreemPage *page;

				handled = TRUE;
				site = screem_application_get_default_site( app );
				page = screem_page_new();
				screem_page_set_pathname( page, url );
				
				/* add document to the site */
				if( ! screem_site_add_page( site, page ) ) {
					/* already open in the site, FIXME: what now? */
				}
				
				/* now display the document in the current view */
				if( screem_window_get_current( window ) != site ) {
					screem_window_set_current( window, site );
				}
				screem_window_set_document( window, page );
			}
			
			gnome_vfs_uri_unref( uri );
		}
		g_list_free( uris );
		break;
	default:
		/* we don't know about this type, and as such
		   this case should never execute */
		screem_window_show_error( SCREEM_WINDOW( window ),
					  _("Error: invalid drop type for window occured\n") );
		break;
	}

	g_free( text );

	gtk_drag_finish( context, handled, FALSE, time );
}


/* GObject Stuff */

#define PARENT_TYPE BONOBO_TYPE_WINDOW

static gpointer parent_class;

static void
screem_window_initialise_class( ScreemWindowClass *klass)
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = G_OBJECT_CLASS( klass );
	widget_class = (GtkWidgetClass *)klass;
	parent_class = g_type_class_peek_parent( klass );

	object_class->finalize = screem_window_finalize;
	object_class->get_property = screem_window_get_prop;
	object_class->set_property = screem_window_set_prop;

	widget_class->show = screem_window_show;

	g_object_class_install_property(object_class,
					ARG_APP_ID,
					g_param_spec_string("app_id",
							    "Application ID",
							    "The ID",
							    "",
							    G_PARAM_READWRITE)
					);

	g_object_class_install_property(object_class,
					ARG_APP,
					g_param_spec_object("app",
							    "Application",
							    "The Application",
							    G_TYPE_OBJECT,
							    G_PARAM_READWRITE |
							    G_PARAM_CONSTRUCT)
					);

	widget_class->realize = screem_window_realize;
	widget_class->size_request = screem_window_size_request;
}

static void
screem_window_initialise( ScreemWindow *window )
{
	window->details = g_new0( ScreemWindowDetails, 1 );

	/* as we only ever open 1 window at the moment we
	   don't need to gen a unique one yet */
	window->details->name = g_strdup( "fixme" );

	window->details->spell = screem_spell_new();
	/* hookup the highlight signal so we can scroll the editor to
	   show it */
	g_signal_connect( G_OBJECT( window->details->spell ),
			  "highlight",
			  G_CALLBACK( screem_window_highlight_spelling ),
			  window );

	window->details->cvs = screem_cvs_new( window );

	gtk_drag_dest_set( GTK_WIDGET( window ),
			   GTK_DEST_DEFAULT_MOTION | 
			   GTK_DEST_DEFAULT_HIGHLIGHT |
			   GTK_DEST_DEFAULT_DROP,
			   drop_types, num_drop_types,
			   GDK_ACTION_COPY );
	
	g_signal_connect( G_OBJECT( window ), "drag_data_received",
			  G_CALLBACK( screem_window_drop ), window );


}

static void
screem_window_set_prop( GObject *object, guint prop_id, 
			const GValue *value, GParamSpec *spec )
{
	const gchar *name;
	gchar *old_name;
	GObject *obj;
	ScreemWindow *window;
	gint w;
	gint h;


	window = SCREEM_WINDOW( object );

	switch( prop_id ) {
	case ARG_APP_ID:
		if( ! G_VALUE_HOLDS_STRING( value ) )
			return;
		name = g_value_get_string( value );
		old_name = bonobo_window_get_name( BONOBO_WINDOW( window ) );
		bonobo_window_set_name( BONOBO_WINDOW( window ),
					name );
		screem_window_constructed( window );
		g_free( old_name );
		break;
	case ARG_APP:
		obj = g_value_get_object( value );
		window->application = SCREEM_APPLICATION( obj );
		w = gconf_client_get_int( window->application->client,
					  "/apps/screem/ui/window_width", NULL );
		h = gconf_client_get_int( window->application->client,
					  "/apps/screem/ui/window_height", NULL );
		
		gtk_window_set_default_size( GTK_WINDOW( window ), w, h );
		break;
	}
}

static void
screem_window_get_prop( GObject *object, guint prop_id, 
			GValue *value, GParamSpec *spec )
{
	GObject *obj;
	gchar *name;

	switch( prop_id ) {
	case ARG_APP_ID:
		name = bonobo_window_get_name( BONOBO_WINDOW( object ) );
		g_value_set_string( value, name );
		g_free( name );
		break;
	case ARG_APP:
		obj = G_OBJECT( SCREEM_WINDOW( object )->application );
		g_value_set_object( value, obj );
		break;
	}
}

static void
screem_window_finalize( GObject *object )
{
	ScreemWindow *window;
	ScreemWindowDetails *details;

	window = SCREEM_WINDOW( object );
	details = window->details;

	if( details->name ) {
		g_free( details->name );
	}

	gconf_client_notify_remove( window->application->client, 
				    details->bar_notify );

	/* remove initial reference */
	g_object_unref( details->tagtree );

	g_object_unref( details->editor );
	g_object_unref( details->preview );
	g_object_unref( details->link_view );
	g_object_unref( details->tree_view );
	
	g_object_unref( details->cvs );
	g_object_unref( details->spell );

	g_object_unref( details->recent_pages );
	g_object_unref( details->recent_sites );

	g_free( details );

	G_OBJECT_CLASS( parent_class )->finalize( object );
}


static void
screem_window_realize( GtkWidget *widget )
{

	/* Create GdkWindow */
	GTK_WIDGET_CLASS( parent_class )->realize( widget );
}

static void
screem_window_size_request( GtkWidget *widget, GtkRequisition *req )
{
	g_return_if_fail( SCREEM_IS_WINDOW( widget ) );
	g_return_if_fail( req != NULL );

	GTK_WIDGET_CLASS( parent_class )->size_request( widget, req );
}

static void
screem_window_show( GtkWidget *widget )
{
	ScreemWindow *window;

	window = SCREEM_WINDOW( widget );

	if( gconf_client_get_bool( window->application->client,
				   "/apps/screem/ui/show_shortcut", NULL ) ) {
		gtk_widget_show( window->details->barbox );
	} else {
		gtk_widget_hide( window->details->barbox );
	}

	GTK_WIDGET_CLASS( parent_class )->show( widget );
}

GType screem_window_get_type()
{
	static GType type = 0;

	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemWindowClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_window_initialise_class,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemWindow ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_window_initialise
		};

		type = g_type_register_static( PARENT_TYPE,
					       "ScreemWindow",
					       &info, 0 );
	}

	return type;
}
